/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/* #define KERBEROS  - uncomment this if you want to try using kerberos */

#define LDAP_PORT	6666
#define LDAP_VERSION	1

#define LDAP_MAX_ATTR_LEN	100

extern int	ldap_dosyslog;

/* debugging stuff */
#ifdef LDAP_DEBUG
extern int	ldap_debug;
#define LDAP_DEBUG_TRACE	0x01
#define LDAP_DEBUG_PACKETS	0x02
#define LDAP_DEBUG_ARGS		0x04
#define LDAP_DEBUG_ANY		0xff

#define Debug( level, fmt, arg1, arg2, arg3 )	if ( ldap_debug & level ) \
		fprintf( stderr, fmt, arg1, arg2, arg3 );
#else
#define Debug( level, fmt, arg1, arg2, arg3 )
#endif

/* 
 * specific LDAP instantiations of BER types we know about
 */

/* general stuff */
#define LDAP_TAG_MESSAGE	0x10
#define LDAP_TAG_MSGID		0x02

/* possible operations a client can invoke */
#define LDAP_REQ_BIND			0x00
#define LDAP_REQ_UNBIND			0x02
#define LDAP_REQ_SEARCH			0x03
#define LDAP_REQ_MODIFY			0x06
#define LDAP_REQ_ADD			0x08
#define LDAP_REQ_DELETE			0x0a
#define LDAP_REQ_MODRDN			0x0c
#define LDAP_REQ_COMPARE		0x0e
#define LDAP_REQ_ABANDON		0x10

/* possible result types a server can return */
#define LDAP_RES_BIND			0x01
#define LDAP_RES_SEARCH_ENTRY		0x04
#define LDAP_RES_SEARCH_RESULT		0x05
#define LDAP_RES_MODIFY			0x07
#define LDAP_RES_ADD			0x09
#define LDAP_RES_DELETE			0x0b
#define LDAP_RES_MODRDN			0x0d
#define LDAP_RES_COMPARE		0x0f
#define LDAP_RES_ANY			(-1)

/* authentication methods available */
#define LDAP_AUTH_SIMPLE	0x00
#define LDAP_AUTH_KRBV4		0x01
#define LDAP_AUTH_KRBV42	0x02

/* filter types */
#define LDAP_FILTER_AND		0x00
#define LDAP_FILTER_OR		0x01
#define LDAP_FILTER_NOT		0x02
#define LDAP_FILTER_EQUALITY	0x03
#define LDAP_FILTER_SUBSTRINGS	0x04
#define LDAP_FILTER_GE		0x05
#define LDAP_FILTER_LE		0x06
#define LDAP_FILTER_PRESENT	0x07
#define LDAP_FILTER_APPROX	0x08

/* substring filter component types */
#define LDAP_SUBSTRING_INITIAL	0x00
#define LDAP_SUBSTRING_ANY	0x01
#define LDAP_SUBSTRING_FINAL	0x02

/* search scopes */
#define LDAP_SCOPE_BASE		0x00
#define LDAP_SCOPE_ONELEVEL	0x01
#define LDAP_SCOPE_SUBTREE	0x02

/* for modifications */
typedef struct ldapmod {
	int		mod_op;
#define LDAP_MOD_ADD		0x00
#define LDAP_MOD_DELETE		0x01
#define LDAP_MOD_REPLACE	0x02
	char		*mod_type;
	char		**mod_values;
	struct ldapmod	*mod_next;
} LDAPMod;

/* 
 * possible error codes we can return
 */

#define LDAP_SUCCESS			0x00
#define LDAP_OPERATIONS_ERROR		0x01
#define LDAP_PROTOCOL_ERROR		0x02
#define LDAP_TIMELIMIT_EXCEEDED		0x03
#define LDAP_SIZELIMIT_EXCEEDED		0x04
#define LDAP_COMPARE_FALSE		0x05
#define LDAP_COMPARE_TRUE		0x06
#define LDAP_STRONG_AUTH_NOT_SUPPORTED	0x07
#define LDAP_STRONG_AUTH_REQUIRED	0x08

#define LDAP_NO_SUCH_ATTRIBUTE		0x10
#define LDAP_UNDEFINED_TYPE		0x11
#define LDAP_INAPPROPRIATE_MATCHING	0x12
#define LDAP_CONSTRAINT_VIOLATION	0x13
#define LDAP_TYPE_OR_VALUE_EXISTS	0x14
#define LDAP_INVALID_SYNTAX		0x15

#define LDAP_NO_SUCH_OBJECT		0x20
#define LDAP_ALIAS_PROBLEM		0x21
#define LDAP_INVALID_DN_SYNTAX		0x22
#define LDAP_IS_LEAF			0x23
#define LDAP_ALIAS_DEREF_PROBLEM	0x24

#define LDAP_INAPPROPRIATE_AUTH		0x30
#define LDAP_INVALID_CREDENTIALS	0x31
#define LDAP_INSUFFICIENT_ACCESS	0x32
#define LDAP_BUSY			0x33
#define LDAP_UNAVAILABLE		0x34
#define LDAP_UNWILLING_TO_PERFORM	0x35
#define LDAP_LOOP_DETECT		0x36

#define LDAP_NAMING_VIOLATION		0x40
#define LDAP_OBJECT_CLASS_VIOLATION	0x41
#define LDAP_NOT_ALLOWED_ON_NONLEAF	0x42
#define LDAP_NOT_ALLOWED_ON_RDN		0x43
#define LDAP_ALREADY_EXISTS		0x44
#define LDAP_NO_OBJECT_CLASS_MODS	0x45

#define LDAP_OTHER			0x50
#define LDAP_SERVER_DOWN		0x51
#define LDAP_LOCAL_ERROR		0x52
#define LDAP_ENCODING_ERROR		0x53
#define LDAP_DECODING_ERROR		0x54
#define LDAP_TIMEOUT			0x55
#define LDAP_AUTH_UNKNOWN		0x56
#define LDAP_FILTER_ERROR		0x57

/*
 * This structure represents both ldap messages and ldap responses.
 * These are really the same, except in the case of search responses,
 * where a response has multiple messages.
 */

typedef struct ldapmsg {
	int		lm_msgid;	/* the message id */
	int		lm_msgtype;	/* the message type */
	BerElement	*lm_ber;	/* the ber encoded message contents */
	struct ldapmsg	*lm_chain;	/* for search - next msg in the resp */
	struct ldapmsg	*lm_next;	/* next response */
} LDAPMessage;
#define NULLMSG	((LDAPMessage *) NULL)

typedef struct ldap {
	Sockbuf		ld_sb;		/* socket descriptor & buffer */
	char		*ld_host;
	int		ld_deref;
#define LDAP_DEREF_NEVER	0
#define LDAP_DEREF_SEARCHING	1
#define LDAP_DEREF_FINDING	2
#define LDAP_DEREF_ALWAYS	3

	int		ld_timelimit;
	int		ld_sizelimit;
#define LDAP_NO_LIMIT		0

	int		ld_errno;
	char		*ld_error;
	int		ld_msgid;

	/* do not mess with these */
	LDAPMessage	*ld_requests;	/* list of outstanding requests */
	LDAPMessage	*ld_responses;	/* list of outstanding responses */
	int		*ld_abandoned;	/* array of abandoned requests */
	char		ld_attrbuffer[LDAP_MAX_ATTR_LEN];
} LDAP;

typedef struct ldaprequest {
	int			lr_msgid;
	int			lr_type;
	struct ldaprequest	*lr_next;
} LDAPRequest;
#define NULLREQ	((LDAPRequest *) NULL)

extern LDAP *ldap_open();
extern LDAPMessage *ldap_first_entry();
extern LDAPMessage *ldap_next_entry();
extern char *ldap_get_dn();
extern char *ldap_dn2ufn();
extern char **ldap_explode_dn();
extern char *ldap_first_attribute();
extern char *ldap_next_attribute();
extern char **ldap_get_values();
extern struct berval **ldap_get_values_len();

#ifdef ultrix
extern char *strdup();
#endif

#ifdef NEEDPROTOS
#include "proto-ldap.h"
#endif NEEDPROTOS
