/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "lber.h"
#include "ldap.h"
#include "common.h"
#include <stdio.h>
#include <string.h>
#include <quipu/commonarg.h>
#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>
#include <quipu/ds_search.h>
#include <quipu/dap2.h>
#include <quipu/dua.h>

do_search( clientsb, m, ber )
Sockbuf		*clientsb;
struct msg	*m;
BerElement	*ber;
{
	int			rc, err;
	int			deref, attrsonly;
	int			sizelimit, timelimit;
	int			msgid = m->m_msgid;
	char			*base;
	char			**attrs;
	struct ds_search_arg	sa;
	static CommonArgs	common = default_common_args;
	extern DN		ldap_str2dn();

	Debug( LDAP_DEBUG_TRACE, "do_search\n", 0, 0, 0 );

	/*
	 * Parse the search request.  It looks like this:
	 *	SearchRequest := [APPLICATION 3] SEQUENCE {
	 *		baseObject	DistinguishedName,
	 *		scope		ENUMERATED {
	 *			baseObject	(0),
	 *			singleLevel	(1),
	 *			wholeSubtree	(2)
	 *		},
	 *		derefAliases	ENUMERATED {
	 *			neverDerefaliases	(0),
	 *			derefInSearching	(1),
	 *			derefFindingBaseObj	(2),
	 *			alwaysDerefAliases	(3)
	 *		},
	 *		sizelimit	INTEGER (0 .. 65535),
	 *		timelimit	INTEGER (0 .. 65535),
	 *		attrsOnly	BOOLEAN,
	 *		filter		Filter,
	 *		attributes	SEQUENCE OF AttributeType
	 *	}
	 */

	if ( ber_scanf( ber, "{aiiiib", &base, &sa.sra_subset, &deref,
	    &sizelimit, &timelimit, &attrsonly ) == -1 ) {
		send_ldap_result( clientsb, LDAP_RES_SEARCH_RESULT, msgid,
		    LDAP_PROTOCOL_ERROR, "" );
		return( 0 );
	}

	sa.sra_baseobject = ldap_str2dn( base );
	free( base );
	if ( sa.sra_baseobject == NULLDN && *base != '\0' ) {
		send_ldap_result( clientsb, LDAP_RES_SEARCH_RESULT, msgid,
		    LDAP_INVALID_DN_SYNTAX, "" );
		return( 0 );
	}

	sa.sra_searchaliases = (deref == 1 || deref == 3);
	if ( deref == 0 || deref == 2 )
		sa.sra_common.ca_servicecontrol.svc_options |=
		    SVC_OPT_DONTDEREFERENCEALIAS;

	sa.sra_common = common;	/* struct copy */
	sa.sra_common.ca_servicecontrol.svc_sizelimit = (sizelimit == 0 ?
	    SVC_NOSIZELIMIT : sizelimit);

	sa.sra_common.ca_servicecontrol.svc_timelimit = (timelimit == 0 ?
	    SVC_NOTIMELIMIT : timelimit);

	sa.sra_eis.eis_infotypes = (attrsonly ? EIS_ATTRIBUTETYPESONLY :
	    EIS_ATTRIBUTESANDVALUES);

	/* search filter */
	if ( (err = get_filter( ber, &sa.sra_filter )) != 0 ) {
		send_ldap_result( clientsb, LDAP_RES_SEARCH_RESULT, msgid,
		    err, "Bad search filter" );
		return( 0 );
	}

#ifdef LDAP_DEBUG
	if ( ldap_debug & LDAP_DEBUG_ARGS ) {
		PS	ps;

		ps = ps_alloc( std_open );
		std_setup( ps, stderr );
		ps_print( ps, "Filter: " );
		fi_print( ps, sa.sra_filter, EDBOUT );
		ps_print( ps, "\n" );
		ps_free( ps );
	}
#endif

	/* attrs to return */
	attrs = NULL;
	if ( ber_scanf( ber, "{v}}", &attrs ) == -1 ) {
		send_ldap_result( clientsb, LDAP_RES_SEARCH_RESULT, msgid,
		    LDAP_PROTOCOL_ERROR, "" );
		return( 0 );
	}
	sa.sra_eis.eis_select = NULLATTR;
	if ( attrs == NULL ) {
		sa.sra_eis.eis_allattributes = 1;
	} else {
		Attr_Sequence	as;
		int		i;

		sa.sra_eis.eis_allattributes = 0;
		for ( i = 0; attrs[i] != NULL; i++ ) {
			as = as_comp_alloc();
			as->attr_acl = NULLACL_INFO;
			as->attr_link = NULLATTR;
			as->attr_value = NULLAV;
			if ( (as->attr_type = AttrT_new( attrs[i] ))
			    == NULLAttrT ) {
				free( as );
				as_free( sa.sra_eis.eis_select );
				send_ldap_result( clientsb,
				    LDAP_RES_SEARCH_RESULT, msgid,
				    LDAP_UNDEFINED_TYPE, "" );
				return( 0 );
			}

			sa.sra_eis.eis_select = as_merge( as,
			    sa.sra_eis.eis_select );
		}
		charlist_free( attrs );
	}

	rc = initiate_dap_operation( OP_SEARCH, m, &sa );

	dn_free( sa.sra_baseobject );
	filter_free( sa.sra_filter );
	as_free( sa.sra_eis.eis_select );

	if ( rc != 0 ) {
		send_ldap_result( clientsb, LDAP_RES_SEARCH_RESULT, msgid,
		    rc, "" );
		return( 0 );
	}

	return( 1 );
}

static get_filter( ber, filt )
BerElement	*ber;
Filter		*filt;
{
	int		tag, len;
	int		err;
	char		typestr[64];
	Filter		f;

	Debug( LDAP_DEBUG_TRACE, "get_filter\n", 0, 0, 0 );

	/*
	 * A filter looks like this coming in:
	 *	Filter ::= CHOICE {
	 *		and		[0]	SET OF Filter,
	 *		or		[1]	SET OF Filter,
	 *		not		[2]	Filter,
	 *		equalityMatch	[3]	AttributeValueAssertion,
	 *		substrings	[4]	SubstringFilter,
	 *		greaterOrEqual	[5]	AttributeValueAssertion,
	 *		lessOrEqual	[6]	AttributeValueAssertion,
	 *		present		[7]	AttributeType,,
	 *		approxMatch	[8]	AttributeValueAssertion
	 *	}
	 */

	f = filter_alloc();
	*filt = f;
	f->flt_next = NULLFILTER;

	switch (tag = ber_peek_tag( ber, &len )) {
	case LDAP_FILTER_EQUALITY:
		Debug( LDAP_DEBUG_ARGS, "EQUALITY\n", 0, 0, 0 );
		f->flt_type = FILTER_ITEM;
		f->FUITEM.fi_type = FILTERITEM_EQUALITY;
		if ( (err = get_ava( ber, &f->FUITEM.UNAVA )) != 0 ) {
			free( f );
			return( err );
		}
		break;

	case LDAP_FILTER_SUBSTRINGS:
		Debug( LDAP_DEBUG_ARGS, "SUBSTRINGS\n", 0, 0, 0 );
		return( get_substring_filter( ber, f ) );
		break;

	case LDAP_FILTER_GE:
		Debug( LDAP_DEBUG_ARGS, "GE\n", 0, 0, 0 );
		f->flt_type = FILTER_ITEM;
		f->FUITEM.fi_type = FILTERITEM_GREATEROREQUAL;
		if ( (err = get_ava( ber, &f->FUITEM.UNAVA )) != 0 ) {
			free( f );
			return( err );
		}
		break;
	case LDAP_FILTER_LE:
		Debug( LDAP_DEBUG_ARGS, "LE\n", 0, 0, 0 );
		f->flt_type = FILTER_ITEM;
		f->FUITEM.fi_type = FILTERITEM_LESSOREQUAL;
		if ( (err = get_ava( ber, &f->FUITEM.UNAVA )) != 0 ) {
			free( f );
			return( err );
		}
		break;

	case LDAP_FILTER_PRESENT:
		Debug( LDAP_DEBUG_ARGS, "PRESENT\n", 0, 0, 0 );
		f->flt_type = FILTER_ITEM;
		f->FUITEM.fi_type = FILTERITEM_PRESENT;
		len = sizeof(typestr);
		if ( ber_scanf( ber, "s", typestr, &len ) == -1 )
			return( LDAP_PROTOCOL_ERROR );
		if ( (f->FUITEM.UNAVA.ava_type = str2AttrT( typestr ))
		    == NULLAttrT )
			return( LDAP_UNDEFINED_TYPE );
		break;

	case LDAP_FILTER_APPROX:
		Debug( LDAP_DEBUG_ARGS, "APPROX\n", 0, 0, 0 );
		f->flt_type = FILTER_ITEM;
		f->FUITEM.fi_type = FILTERITEM_APPROX;
		if ( (err = get_ava( ber, &f->FUITEM.UNAVA )) != 0 ) {
			free( f );
			return( err );
		}
		break;

	case LDAP_FILTER_AND:
		Debug( LDAP_DEBUG_ARGS, "AND\n", 0, 0, 0 );
		f->flt_type = FILTER_AND;
		return( get_filter_list( ber, f ) );
		break;

	case LDAP_FILTER_OR:
		Debug( LDAP_DEBUG_ARGS, "OR\n", 0, 0, 0 );
		f->flt_type = FILTER_OR;
		return( get_filter_list( ber, f ) );
		break;

	case LDAP_FILTER_NOT:
		Debug( LDAP_DEBUG_ARGS, "NOT\n", 0, 0, 0 );
		f->flt_type = FILTER_NOT;
		(void) ber_skip_tag( ber, &len );
		return( get_filter( ber, &f->FUFILT ) );
		break;

	default:
		Debug( LDAP_DEBUG_ANY, "unknown filter type %d\n", tag, 0, 0 );
		free( f );
		return( LDAP_PROTOCOL_ERROR );
		break;
	}

	return( 0 );
}

static get_filter_list( ber, f )
BerElement	*ber;
Filter		f;
{
	Filter	new, tail;
	int	tag, len, err;
	char	*last;

	Debug( LDAP_DEBUG_TRACE, "get_filter_list\n", 0, 0, 0 );

	f->FUFILT = tail = NULLFILTER;
	for ( tag = ber_first_element( ber, &len, &last ); tag != -1;
	    tag = ber_next_element( ber, &len, last ) ) {
		if ( (err = get_filter( ber, &new )) != 0 )
			return( err );

		if ( f->FUFILT == NULLFILTER ) {
			f->FUFILT = new;
		} else {
			tail->flt_next = new;
		}
		tail = new;
	}

	return( 0 );
}

static get_substring_filter( ber, f )
BerElement	*ber;
Filter		f;
{
	int		tag, len;
	char		typestr[64];
	AttributeType	type;
	char		*valstr;
	AttributeValue	value;

	Debug( LDAP_DEBUG_TRACE, "get_substring_filter\n", 0, 0, 0 );

	f->flt_type = FILTER_ITEM;
	f->FUITEM.fi_type = FILTERITEM_SUBSTRINGS;
	len = sizeof(typestr);
	if ( ber_scanf( ber, "{s{", typestr, &len ) == -1 ) {
		return( LDAP_PROTOCOL_ERROR );
	}
	if ( (type = str2AttrT( typestr )) == NULLAttrT ) {
		return( LDAP_UNDEFINED_TYPE );
	}
	f->FUITEM.UNSUB.fi_sub_type = type;
	f->FUITEM.UNSUB.fi_sub_initial = NULLAV;
	f->FUITEM.UNSUB.fi_sub_any = NULLAV;
	f->FUITEM.UNSUB.fi_sub_final = NULLAV;
	while ( (tag = ber_peek_tag( ber, &len )) != -1
	    && tag != LBER_SEQUENCE ) {
		AV_Sequence	avs;

		if ( ber_scanf( ber, "a", &valstr ) == -1 ) {
			free( f );
			return( LDAP_PROTOCOL_ERROR );
		}

		if ( (value = str2AttrV( valstr, type->oa_syntax ))
		    == NULLAttrV ) {
			return( LDAP_INVALID_SYNTAX );
		}

		if ( (avs = avs_comp_new( value )) == NULLAV )
			return( LDAP_OTHER );

		switch ( tag ) {
		case LDAP_SUBSTRING_INITIAL:
			if ( f->FUITEM.UNSUB.fi_sub_initial != NULLAV
			    && f->FUITEM.UNSUB.fi_sub_initial->avseq_next
			    != NULLAV ) {
				return( LDAP_PROTOCOL_ERROR );
			}
			f->FUITEM.UNSUB.fi_sub_initial = avs_merge( avs,
			    f->FUITEM.UNSUB.fi_sub_initial );
			break;

		case LDAP_SUBSTRING_ANY:
			f->FUITEM.UNSUB.fi_sub_any = avs_merge( avs,
			    f->FUITEM.UNSUB.fi_sub_any );
			break;

		case LDAP_SUBSTRING_FINAL:
			if ( f->FUITEM.UNSUB.fi_sub_final != NULLAV
			    && f->FUITEM.UNSUB.fi_sub_final->avseq_next
			    != NULLAV ) {
				return( LDAP_PROTOCOL_ERROR );
			}
			f->FUITEM.UNSUB.fi_sub_final = avs_merge( avs,
			    f->FUITEM.UNSUB.fi_sub_final );
			break;

		default:
			return( LDAP_PROTOCOL_ERROR );
		}
	}

	return( 0 );
}

search_result( sb, msgid, sr )
Sockbuf			*sb;
int			msgid;
struct ds_search_result	*sr;
{
	EntryInfo	*e;
	BerElement	*ber;
	int		rc;

	Debug( LDAP_DEBUG_TRACE, "search_result\n", 0, 0, 0 );

	if ( ! sr->srr_correlated ) {
		Debug( LDAP_DEBUG_ARGS, "correlating results\n", 0, 0, 0 );
		correlate_search_results( sr );
	}

	for ( e = sr->CSR_entries; e != NULLENTRYINFO; e = e->ent_next ) {
		Debug( LDAP_DEBUG_ARGS, "\tentry:\n", 0, 0, 0 );

		if ( (ber = ber_alloc()) == NULLBER ) {
			send_ldap_result( sb, LDAP_RES_SEARCH_RESULT, msgid,
			    LDAP_OPERATIONS_ERROR, "" );
			return;
		}

		if ( ber_printf( ber, "{it{", msgid, LDAP_RES_SEARCH_ENTRY )
		    == -1 )
			return;

		if ( encode_dn( ber, e->ent_dn ) == -1 )
			return;

		if ( encode_attrs( ber, e->ent_attr ) == -1 )
			return;

		if ( ber_printf( ber, "}}" ) == -1 )
			return;

		(void) ber_flush( sb, ber, 1 );
	}

	switch ( sr->CSR_limitproblem ) {
	case LSR_NOLIMITPROBLEM:
		rc = LDAP_SUCCESS;
		break;
	case LSR_TIMELIMITEXCEEDED:
		rc = LDAP_TIMELIMIT_EXCEEDED;
		break;
	case LSR_SIZELIMITEXCEEDED:
	case LSR_ADMINSIZEEXCEEDED:
		rc = LDAP_SIZELIMIT_EXCEEDED;
		break;
	}

	Debug( LDAP_DEBUG_ARGS, "\tresult:\n", 0, 0, 0 );
	send_ldap_result( sb, LDAP_RES_SEARCH_RESULT, msgid, rc, "" );

	return;
}
